/***********************************************************************

Copyright (c) 2002 Fritz Ganter <ganter@ganter.at>

Website: http://gpsdrive.kraftvoll.at


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    *********************************************************************

Speedstep is a daemon for linux kernel using ACPI which controls the
CPU speed on SpeedStep CPUs depentend of the CPU load

$Log: autospeedstep.c,v $
Revision 1.2  2002/07/12 09:23:12  ganter
added manpage

Revision 1.1  2002/07/12 08:16:32  ganter
changed filenames

Revision 1.3  2002/07/12 08:09:27  ganter
changed values

Revision 1.2  2002/07/09 23:19:31  ganter
works fine

Revision 1.1.1.1  2002/07/09 22:37:24  ganter
created


***********************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>


static int downcounter = 0, state = 3;
static const char *progname;
static int daemonize (void);
static const char *logfile = "/var/log/speedstep";
int highval=70;
int lowval=20;

void
cpuup ()
{
  FILE *fp;
  time_t curtime;
  char *timestr;

  if (state == 0)
    return;
  fp = fopen ("/proc/acpi/processor/CPU0/performance", "w");
  if (fp == NULL)
    {
      perror ("/proc/acpi/processor/CPU0/performance");
      exit (2);
    }
  fprintf (fp, "0");
  fclose (fp);
  downcounter = 0;
  curtime = time (NULL);
  timestr = ctime (&curtime);
  timestr[strlen (timestr) - 1] = '\0';
  fprintf (stderr, timestr);
  fprintf (stderr, ": max. speed mode\n");
  state = 0;
}

void
cpudown ()
{
  FILE *fp;
  time_t curtime;
  char *timestr;

  if (state == 1)
    return;
  downcounter++;
  if (downcounter < 5)
    return;
  fp = fopen ("/proc/acpi/processor/CPU0/performance", "w");
  if (fp == NULL)
    {
      perror ("/proc/acpi/processor/CPU0/performance");
      exit (2);
    }
  fprintf (fp, "1");
  fclose (fp);

  curtime = time (NULL);
  timestr = ctime (&curtime);
  timestr[strlen (timestr) - 1] = '\0';
  fprintf (stderr, timestr);
  fprintf (stderr, ": power save mode\n");
  state = 1;
}


int
main (int argc, char *argv[])
{
  int e;
  unsigned int cuse, cice, load, csys, oldcuse = 0;
  unsigned long cide;
  FILE *fp;



  progname = (const char *) strrchr (argv[0], '/');
  progname = progname ? (progname + 1) : argv[0];

  /* check perms */
  if (geteuid () != 0)
    {
      fprintf (stderr, "%s: must be run as root\n", progname);
      exit (1);
    }


  if (daemonize () < 0)
    exit (3);

  while (1)
    {
      fp = fopen ("/proc/stat", "r");
      if (fp == NULL)
	{
	  perror ("/proc/stat");
	  exit (1);
	}

      do
	{
	  e = fscanf (fp, "cpu  %u %u %u %lu", &cuse, &cice, &csys, &cide);
	}
      while (e != 4);
      if (oldcuse != 0)
	{
	  load = cuse +csys - oldcuse;
/* 	  printf ("%u\n", load); */
	  if (load > highval)
	    cpuup ();
	  if (load < lowval)
	    cpudown ();
	}
      oldcuse = cuse +csys;
      fclose (fp);
      usleep (1000000);
    }


  return 0;
}

static int
daemonize (void)
{
  int logfd;
  int nullfd;

  switch (fork ())
    {
    case -1:
      fprintf (stderr, "%s: fork: %s\n", progname, strerror (errno));
      return -1;
    case 0:
      /* child */
      break;
    default:
      /* parent */
      exit (0);
    }

  /* disconnect */
  setsid ();
  umask (0);

  /* set up stdout, stderr to log and stdin to /dev/null */
  nullfd = open ("/dev/null", O_RDONLY, 0640);
  if (nullfd < 0)
    {
      fprintf (stderr, "%s: can't open %s: %s\n", progname,
	       "/dev/null", strerror (errno));
      return -1;
    }
  logfd = open (logfile, O_WRONLY | O_CREAT | O_APPEND);
  if (logfd < 0)
    {
      fprintf (stderr, "%s: can't open %s: %s\n", progname,
	       logfile, strerror (errno));
      return -1;
    }

  if (dup2 (logfd, STDIN_FILENO) != STDIN_FILENO)
    {
      fprintf (stderr, "%s: dup2: %s\n", progname, strerror (errno));
      return -1;
    }
  if (dup2 (logfd, STDOUT_FILENO) != STDOUT_FILENO)
    {
      fprintf (stderr, "%s: dup2: %s\n", progname, strerror (errno));
      return -1;
    }
  if (dup2 (logfd, STDERR_FILENO) != STDERR_FILENO)
    {
      fprintf (stderr, "%s: dup2: %s\n", progname, strerror (errno));
      return -1;
    }

  /* get outta the way */
  chdir ("/");

  return 0;
}
